#!/bin/zsh -f

# build.zsh,v 1.11 2002/07/22 16:36:01 tril Exp
# Junk script with which to build Tunes.
#
# build.zsh,v
# Revision 1.11  2002/07/22 16:36:01  tril
# added support to the build system for the members php
#
# Revision 1.10  2001/09/04 00:28:42  dem
# New footers:
# * date last changed
# * author of last change
# * cvsweb link for every page
# * webmaster link is in the html, no longer preprocessed
#
# Revision 1.9  2001/07/03 15:54:10  dem
# commented cvs IDs and added a footer to phtml processing
#
# Revision 1.8  2000/02/10 15:01:47  fare
# Cleaner cleaning (don't delete CVS/ stuff).
# Still don't include CVS stuff in .tar.bz2 archive, either (or should we?).
#
# Revision 1.7  1999/11/18 17:33:18  fare
# Clarifying the license: we're using GNU LGPL v2.1 or later.
#
# Revision 1.6  1999/06/02 23:28:20  fare
# Commented out precambrian tunes file exclusion (.map and stuff),
# so as to allow retro keymaps to be included in snapshot.
#
# Revision 1.5  1999/04/05 23:13:00  fare
# Allowing for snapshots of both tunes and retro
#
# Revision 1.4  1999/03/13 12:12:54  fare
# *** empty log message ***
#
# Revision 1.3  1998/11/27 01:06:20  fare
# surface corrections in the web pages.
#
# Revision 1.2  1998/11/26 05:25:23  fare
# ensure_dir not just when making docs.
#
# Revision 1.1.1.1  1998/11/26 04:10:47  tunes
# Imported reshaped, post-0.0.0.36 sources
#
#

setopt EXTENDED_GLOB NO_NOMATCH NUMERIC_GLOB_SORT BRACE_CCL RC_QUOTES
setopt MAGIC_EQUAL_SUBST


##################### READ THE SETUP FROM RESOURCE FILE #####################

# There can be a .tunesrc file in your home directory, that should contain
# the following definitions, along with any compatibility kludge
# (zsh options, function definitions, PATH adjustment, security checking, etc):
#
# TUNES		path for the top of the Tunes development tree
# WWW		path under which to make WWW archives files
# TMP		path for a relatively trusted temporary directory
# ARCHIVETYPE	Your preferred extension for Tunes archives {tgz,tbz,tb2,zip}
# LOGFILE	A file where to log tunes activity
# BACKUPDIR	a place where to store backups of files being modified
#
# My own tunesrc is given as an example in $TUNES/bin/tunesrc.bespin
#
# If no such file is present, values will be guessed, at your own risks.
#

find_rcfile () {
    local i
    for i in $TUNESRC $HOME/.tunesrc /etc/tunesrc ; do
	if [ -f $i ] ; then
	    TUNESRC=$i ; return 0
	fi
    done
    return 1
}
find_status () {
    if find_rcfile ; then
	TUNES=
	source $TUNESRC
	if [ ! -d $TUNES ] ; then
	    abort 102 "Bad Tunes status file -- check $TUNESRC"
	fi
    else
	guess_status
	if [ ! -d $TUNES ] ; then
	    abort 102 "Could not guess status."
	fi
    fi
    : ${TMP:=/tmp}
    : ${ARCHIVETYPE:=tgz}
}
guess_status () {
  echo >& 2 "No \$TUNESRC or .tunesrc file found; Guessing status..."
  : ${TUNES:=$HOME/tunes}
  : ${WWW:=$TUNES/html}
  : ${ARCHIVETYPE:=tb2}
  : ${TMP:=/tmp}
  : ${BACKUPDIR:=/tmp}
  : ${LOGFILE:=$TMP/tunes-$HOST-$UID-$$.log}
}

LF='
'




############################## A FEW UTILITIES ###############################

abort () {
  CODE=$1 ; shift
  [ -n "$*" ] && print -r >&2 "$*"
  exit $CODE
}

println () {
  for i ; do print -r "$i" ; done
}

bztar () {
  tar --use-compress-program bzip $@
}
b2tar () {
  tar --use-compress-program bzip2 $@
}
unarc () {
  case "$1" in
    *.tgz|*.tar.gz) tar zxvf $@ ;;
    *.tbz|*.tar.bz) bztar -xvf $@ ;;
    *.tb2|*.tar.bz2) b2tar -xvf $@ ;;
    *.zip|*.ZIP) unzip $@ ;;
  esac
}

newdir () { mkdir $1 && cd $1 || abort 3 "Cannot newdir $1" }



function_p () {
  local WHENCE="$(whence -v $1)"
  case "$WHENCE" in
    "$1 is a shell function") return 0 ;; *) return 1 ;;
  esac
}
command_p () { local WHENCE="$(whence -p $1)" }
call_if_func () {
  if function_p "$1" ; then
    $@
  else
    abort 102 $NO_FUNC_MSG
  fi
}

ensure_dir () {
  [ -d $1 ] || mkdir -p $1 # GNU mkdir
}
ensure_dir () {
  local DIR=$(dirname $1) BASE=$(basename $1)
  [ -d $1 ] && return
  ensure_dir $DIR && ( cd $DIR && mkdir $BASE )
}


fullpath () { case "$1" in /*) echo $1 ;; *) echo $PWD/$1 ;; esac }


repl () {(
  eval "prog=($1)" ; shift
  eval "FILES=($1)" ; shift
  for i in $FILES ; do
    j="${BACKUPDIR}/$(fullpath $i | perl -npe 's,/,!,g ; chop' )~" &&
    $prog $@ < $i > $j &&
    (cmp -s $j $i || =cp $j $i)
  done
  # [[ -f "$j" ]] && =rm -f $j
)}
replperl () { repl perl $@ }
replsed () { repl sed $@ }


READ_MSG () {
  : READ_MSG
  setopt LOCAL_OPTIONS NO_XTRACE
  MSG=$(cat)
}

PRINT_MSG () {
  : PRINT_MSG
  setopt LOCAL_OPTIONS NO_XTRACE
  print -r $MSG
}

######################## WORKING WITH EXTERNAL PROGRAMS ######################
#
# Various hacks to keep the sources "clean".
# Run what you need before committing changes...
#

detect_sgmltools () {
  if SGML2TXT=$(whence -p sgml2txt) ; then
    # Lowly hack to detect sgmltools 0.99
    if head -n 4 < $SGML2TXT | grep '^..Greg Hankins' > /dev/null ; then
      echo 0.x
    else
      echo 1.x
    fi
  fi
}

setup_sgmltools () {
  case $(detect_sgmltools) in
    "")
      SGMLTOOLS=
    ;;
    0.*)
      SGMLTOOLS=y
      SGMLFLAGS="-l"
    ;;
    1.*)
      SGMLTOOLS=y
      SGMLFLAGS="-c latin"
    ;;
  esac
  if [ -n "$SGMLTOOLS" ] ; then
    SGML2TXT==sgml2txt
    SGML2HTML==sgml2html
    SGML2LATEX==sgml2latex
  fi
}


############################## CLEANING UP ###################################

CLEAN_PAT=(
	   -name core		# shouldn't happen, but who knows?
	-o -name '*.html'	# .html is generated files! source is .phtml
	-o -name '*.orig'
	-o -name '*.rej'
	-o -name '*.bak'
	-o -name '*.config'
	-o -name '*.[oasilbxpzZ]'
	-o -name '*.[zv]?'
	-o -name '*.cm?'
	-o -name '*.m4f'
	-o -name '*.obj'
	-o -name '*.bin'
	-o -name '*.exe'
	#-o -name '*.co[m12]'
	#-o -name '*.[cx]rf'
	#-o -name '*.map'
	-o -name '*.lst'
	-o -name '*.$$$'
	-o -name '*.tmp'
	-o -name '*.dvi'	# LaTeX DeViceIndependent output
	-o -name '*.ps'		# PostScript
	-o -name '*.log'	# LaTeX log
	-o -name '*.aux'	# LaTeX auxiliary build file
	-o -name '*.blg'	# BibTeX log
	-o -name '*.bbl'	# BibTeX auxiliary build file
	-o -name '*.toc'	# LaTeX table of contents
	-o -name '*.haux'	# hevea auxiliary build file
	-o -name '*~'		# EMACS backup
	-o -name '#*#'		# EMACS emergency backup
	-o -name foo		# FOO...
	-o -name 'ba[rz]'
	-o -name 'qu*ux'
	-o -name 'z?p'
	-o -name 't?t?'		# ... BAR!
	-o -name '*.scm-*'	# intermediate Scheme source
	-o -name 'compiled'	# compiler output
	-o -name 'compiled.*'
	-o -name '.tmp*'
	-o -name 'patch.*'
	-o -name 'version.h'
	# Additional Retro stuff -- http://retro.tunes.org/
	-o -name image
	-o -name modules
	-o -name writeboot
	-o -name modutil
	-o -name '*.rdf'
)

VERYCLEAN_PAT=(
	$CLEAN_PAT
	-o -name ".depend"
	-o -name "config.sh"
	-o -name "*.tar*"
	-o -name "*.tb2"
	-o -name "*.t?z"
)
MRPROPER_PAT=(
	$VERYCLEAN_PAT
)

clean () {
   BAD="`find . '(' $CLEAN_PAT ')' -print`"
   echo $BAD >&2
   /bin/rm -f ${=BAD}
}
veryclean () {
   BAD="`find . '(' $VERYCLEAN_PAT ')' -print`"
   echo $BAD >&2
   /bin/rm -f ${=BAD}
}
mrproper () {
   BAD="`find . '(' $MRPROPER_PAT ')' -print`"
   echo $BAD >&2
   /bin/rm -f ${=BAD}
}

############################# DEPENDENCIES ###################################
#
# Computing make(1) dependencies for the TUNES project...
#
# In any given compilation process, we distinguish
# a) the source(s) from which to process
# b) the host(s) where to store the processed information
# c) the destination(s) in which the processed files are meant to run
# The equation is that the abstract meaning of the source
# is the same as the result of evaluating the hosted result
# in the destination environment.
#
# This separation means that we have cleaner metaprogramming in general
# and cross-compilation in particular.
# Also, by having object files made in a non-source directory,
# we avoid accidentally adding cruft to the CVS repository,
# and allow for (sequentially) multi-config builds.
# As much as possible, everything should be done in a declarative way,
# and information should be propagated automatically.
# Manual redundancy maintenance is EVIL.
#
# Possible Improvements:
# * make it possible that anything be built from anywhere, not just its
#  source directory, without recursive calls to make.
# * make parallel multi-config builds possible, as well as making
#  from target directory, etc.
# * get rid from make(1) and have it all managed from within Tunes.
#

depend () {

### If we're lucky, the directory needs nothing to be done.
if [ ! -f "Makefile" ] ; then
  echo "no Makefile, no dependencies" >& 2
  return 0
fi

### Define auxiliary functions
list_sources () {
  # Source files with known suffixes have known rules.
  # To avoid rules, use another suffix,
  # or put files in a directory without Makefile.
  setopt LOCAL_OPTIONS NULL_GLOB NO_MATCH
  SOURCES=( *.sgml *.png *.jpg *.cat *.php *.inc # *.phtml
		$([ ! -f SOURCES ] || cat SOURCES) )
}
list_subdirs () {
  # All subdirectories are to be processed, recursively;
  # this stops if the subdirectory has no Makefile (see above).
  setopt LOCAL_OPTIONS NULL_GLOB NO_MATCH
  SUBDIRS=( *(/) )
}
get_dirs () {
  # Detect the source directory, the host directory, etc.
  SRCCUR=$(pwd -r)/
  SRCTOP=$(cd ${TUNES}/src/;pwd -r)
  SRCSUB=${SRCCUR#$SRCTOP/}
  HOSTOP=${WWW}/
  SOURCE_DIR=
  HOST_DIR=${HOSTOP}${SRCSUB}
}
rule () {
# dump a Makefile rule
    local COMMAND
    echo "$1:	$2" ; shift 2
    for COMMAND ; do echo "	$COMMAND" ; done
    echo
}
f_rule () {
# rule: use target name in source directory to produce file in host directory
    local S="${SOURCE_DIR}$1" H="${HOST_DIR}$1"
    echo ".PHONY: $S"
    rule "$S" "$H"
}

id_rule () {
# rule: just copy a file as is in new location
    local S="${SOURCE_DIR}$1" H="${HOST_DIR}$2"
    rule "$H" "$S" "\$(CP) $S $H"
}

#p_rule () {
## rule: preprocess file from source location to host one.
#    local P="$1" S="${SOURCE_DIR}$2" H="${HOST_DIR}$3"
#    f_rule "$3"
#    rule "$H" "$S" "$1 < $S > $H"
#}

begin_rules () {
  BASENAME="${FILE%.*}"
  HOSTFILE=${HOST_DIR}$FILE
  SRC=${SOURCE_DIR}$FILE
  HOSTBASE=${HOST_DIR}$BASENAME
  BFILES=""
  DFILES=""
  BDIRS=""
  DDIRS=""
  SDIRS=""
}
end_rules () {
  ALLBFILES="$ALLBFILES $BFILES"
  ALLDFILES="$ALLDFILES $DFILES"
  ALLBDIRS="$ALLBDIRS $BDIRS"
  ALLDDIRS="$ALLDDIRS $DDIRS"
  ALLSDIRS="$ALLSDIRS $SDIRS"
}
standard_file_rules () {
  case $FILE in
    *.html)
      id_rule $FILE
      DFILES="$HOSTFILE"
      ;;
#    *.phtml)
#      p_rule $PHTML $BASENAME.phtml $BASENAME.html
#      DFILES="$HOSTBASE.html"
#      ;;
    *.txt|*.gif|*.jpg|*.png)
      id_rule $FILE $FILE
      DFILES="$HOSTFILE"
      ;;
    *.cat)
      id_rule $FILE $BASENAME
      DFILES="$HOSTBASE"
      ;;
    *.php|*.inc)
      id_rule $FILE $FILE
      DFILES="$HOSTFILE"
      ;;
    *.sgml)
      id_rule $FILE $FILE
      base=$(basename $FILE .sgml)
      if [ -n "$HOST_DIR" ] ; then
        CD="cd $HOST_DIR"
      else
        CD=":"
      fi
      if [ -n "$SGMLTOOLS" ] ; then
        f_rule "$BASENAME.html"
        rule "$HOSTBASE.html" "$HOSTFILE" "$CD ; $SGML2HTML $SGMLFLAGS $FILE"
        f_rule "$BASENAME.txt"
        rule "$HOSTBASE.txt" "$HOSTFILE" "$CD ; $SGML2TXT $SGMLFLAGS $FILE"
        f_rule "$BASENAME.dvi"
        rule "$HOSTBASE.dvi" "$HOSTFILE" "$CD ; $SGML2LATEX $SGMLFLAGS $FILE"
        f_rule "$BASENAME.ps"
        rule "$HOSTBASE.ps" "$HOSTBASE.dvi" \
			    "$CD ; dvips -o $BASENAME.ps $BASENAME.dvi"
        DFILES="$HOSTFILE $HOSTBASE.html $HOSTBASE.txt"
        # DFILES="$HOSTFILE $HOSTBASE.dvi $HOSTBASE.ps"
      fi
      ;;
    *)
      echo "# Warning: can't handle source file $FILE"
      echo "# Warning: can't handle source file $FILE" >&2
      ;;
  esac
}
standard_dir_rules () {
  pushd $DIR
  make depend >&2
  if grep "^# Has documentation" .depend >&/dev/null ; then
     DDIRS=$DIR-doc
     SDIRS=$DIR
  fi
  if grep "^# Has binaries" .depend >&/dev/null ; then
     BDIRS=$DIR-bin
     SDIRS=$DIR
  fi
  popd
  rule ".PHONY" "$DIR $DIR-doc $DIR-both $DIR-bin $DIR-doc"
  rule "$DIR-doc" "" \
	"\$(BUILD) ensure_dir ${HOST_DIR}$DIR" \
	"make -C $DIR doc"
  rule "$DIR-bin" "" \
	"make -C $DIR bin"
  rule "$DIR-both" "" \
	"\$(BUILD) ensure_dir ${HOST_DIR}$DIR" \
	"make -C $DIR both"
  rule "$DIR" "" \
	"\$(BUILD) ensure_dir ${HOST_DIR}$DIR" \
	"make -C $DIR"
}


### Get parameters
setup_sgmltools
get_dirs
#PHTML="\$(BUILD) process_phtml"		# preprocessor for .phtml files

### Initialize variables
ALLBFILES=		# list of binary files to be made
ALLDFILES=		# list of documentation files to be made
ALLBDIRS=		# list of directories with bin files to be made
ALLDDIRS=		# list of directories with doc files to be made
ALLSDIRS=		# list of directories with something to be made

### Rules
list_sources
for FILE in $SOURCES ; do
  begin_rules
  standard_file_rules
  end_rules
done
list_subdirs
for DIR in $SUBDIRS ; do
  if [ -f $DIR/Makefile ] ; then
  begin_rules
  standard_dir_rules
  end_rules
  fi
done
if [ -f depend.zsh ] ; then
  source depend.zsh
fi

### Global rules
rule .PHONY "subdoc subbin subboth sub heredoc herebin"

rule subdoc "$ALLDDIRS"
rule subbin "$ALLBDIRS"
rule subboth "$ALLSDIRS"
rule heredoc "$ALLDFILES"
rule herebin "$ALLBFILES"
rule sub subboth
rule hereboth "herebin heredoc"
rule documentation "heredoc subdoc"
rule binaries "herebin subbin"
rule both "hereboth subboth"

[ -n "$ALLDDIRS$ALLDFILES" ] && echo "# Has documentation"
[ -n "$ALLBDIRS$ALLBFILES" ] && echo "# Has binaries"

#rule toto: test "echo toto"

### Done
return 0
}

wdepend () {
   abort "Burp!"
}


############################ SOURCE FIXUP ###################################
#
# Various hacks to keep the sources "clean".
# Run what you need before committing changes...
#

remove_silly_html_markups () {
  (cd $TUNES/doc/ &&
   replperl '**/*.html' -npe '
     s/\Q<!-- Changed by:\E[^>]*\>\n// ;
    '
  )
}


############################# HTML PREPROCESSING ############################
#
# Let's start it simple
#

#process_phtml () {
#    # Remove "end tags"
#    sed 's,</body>,,i' | sed 's,</html>,,i'
#}

############################## MAKING AN ARCHIVE ############################
#
find_good_files () {
    find $@ -type f -a '!' '(' -path '*/CVS/*' -o $MRPROPER_PAT ')' -print
}

make_archive () {
    : ${ARCHIVE_PROJECT:=${1:-tunes}}
    : ${ARCHIVE_FILE:=../${ARCHIVE_PROJECT}-$(date +%Y%m%d).tar.bz2}
    rm -f $ARCHIVE_PROJECT
    ln -s . $ARCHIVE_PROJECT
    b2tar -cvf ${ARCHIVE_FILE} $(find_good_files $ARCHIVE_PROJECT/)
    rm -f $ARCHIVE_PROJECT
}

# Same for retro
make_retro_archive () {
    : ${ARCHIVE_FILE:=../retro-$(date +%Y%m%d).tar.bz2}
    rm -f retro
    ln -s . retro
    b2tar -cvf ${ARCHIVE_FILE} $(find_good_files retro/)
    rm -f retro
}

################################# LOGGING ###################################
#
# Use `TUNESADM LOG foo...` instead of `TUNESADM foo...`
# to have everything logged to your usual LOGFILE
#

LOG () {
  echo >> $LOGFILE &&
  ( echo "$(date): TUNESADM LOG $*" ;
    DEBUGGING=y
    restore-x
    call_if_func $@
  ) >>& $LOGFILE
}

############################### MAIN PROGRAM ################################

find_status

UTSL="Use The Source, Luke!"

[ -n "$1" ] || abort 1 "TUNES build.sh usage: $UTSL"

NO_FUNC_MSG="Unknown function $1 -- Valid functions: $UTSL"
call_if_func $@
